import time

from nyckel import ClassificationAnnotation, ClassificationLabel, TextClassificationFunction, TextClassificationSample


def test_simple(text_classification_function_with_content: TextClassificationFunction) -> None:
    func: TextClassificationFunction = text_classification_function_with_content

    samples = func.list_samples()
    assert len(samples) > 0
    func.delete_samples([sample.id for sample in samples if sample.id])
    time.sleep(2)
    samples = func.list_samples()
    assert len(samples) == 0


def test_duplicate_samples(text_classification_function: TextClassificationFunction) -> None:
    func: TextClassificationFunction = text_classification_function

    samples = [
        TextClassificationSample(data="hello"),
        TextClassificationSample(data="hello"),
    ]

    sample_ids = func.create_samples(samples)
    assert len(sample_ids) == 1  # Duplicate samples are ignored when posting.


def test_update_annotation(text_classification_function: TextClassificationFunction) -> None:
    func: TextClassificationFunction = text_classification_function

    labels_to_create = [ClassificationLabel(name="Nice"), ClassificationLabel(name="Bad")]
    nice = ClassificationAnnotation(label_name="Nice")
    bad = ClassificationAnnotation(label_name="Bad")
    sample = TextClassificationSample(data="Hello, friend", external_id="1", annotation=nice)

    # Create assets and make sure the annotation is set correctly.
    func.create_labels(labels_to_create)
    sample_id = func.create_samples([sample])[0]
    time.sleep(1)
    sample = func.read_sample(sample_id)
    assert sample.annotation and sample.annotation.label_name == "Nice"

    # Remove the annotation, update it and make sure it's updated.
    sample.annotation = None
    func.update_annotation(sample)
    time.sleep(1)
    sample = func.read_sample(sample_id)
    assert not sample.annotation

    # Change the annotation, update it and make sure it's updated.
    sample.annotation = bad
    func.update_annotation(sample)
    time.sleep(1)
    sample = func.read_sample(sample_id)
    assert sample.annotation and sample.annotation.label_name == "Bad"



import time

import requests


class OAuth2Renewer:
    def __init__(self, client_id: str, client_secret: str, server_url: str = "https://www.nyckel.com"):
        self._client_id = client_id
        self._client_secret = client_secret
        self._server_url = server_url.rstrip("/")
        self._renew_at = 0
        self._bearer_token: str = ""

    @property
    def client_id(self) -> str:
        return self._client_id

    @property
    def server_url(self) -> str:
        return self._server_url

    @property
    def token(self) -> str:
        if time.time() > self._renew_at:
            self._renew_token()
        return self._bearer_token

    def _renew_token(self) -> None:
        RENEW_MARGIN_SECONDS = 10 * 60

        token_url = f"{self._server_url}/connect/token"
        data = {
            "client_id": self._client_id,
            "client_secret": self._client_secret,
            "grant_type": "client_credentials",
        }

        response = requests.post(token_url, data=data)
        if not response.status_code == 200:
            raise ValueError(f"{response.status_code=} Failed to renew credentials at {token_url=} using {data=}.")

        self._bearer_token = response.json()["access_token"]
        self._renew_at = time.time() + response.json()["expires_in"] - RENEW_MARGIN_SECONDS



import time

from nyckel import (
    ClassificationAnnotation,
    ClassificationLabel,
    OAuth2Renewer,
    TextClassificationFunction,
    TextClassificationSample,
)
from nyckel.functions.classification.classification import ClassificationFunction
from nyckel.orchestration import NyckelLabelDeleter


def hold_until_list_samples_available(function: ClassificationFunction, expected_count: int) -> None:
    actual_count = 0
    while not actual_count == expected_count:
        actual_count = len(function.list_samples())
        time.sleep(0.25)


def test_simple(text_classification_function: TextClassificationFunction, auth_test_user: OAuth2Renewer) -> None:
    func: TextClassificationFunction = text_classification_function

    labels_to_create = [ClassificationLabel(name="Nice"), ClassificationLabel(name="Boo")]
    func.create_labels(labels_to_create)
    nice = ClassificationAnnotation(label_name="Nice")
    boo = ClassificationAnnotation(label_name="Boo")
    samples = [
        TextClassificationSample(data="hello", external_id="1", annotation=nice),
        TextClassificationSample(data="hi", external_id="2", annotation=nice),
        TextClassificationSample(data="Good bye", external_id="3", annotation=boo),
        TextClassificationSample(data="I'm leaving", external_id="4", annotation=boo),
        TextClassificationSample(data="Hi again", external_id="5"),
    ]
    func.create_samples(samples)
    hold_until_list_samples_available(func, 5)

    deleter = NyckelLabelDeleter(func.function_id, auth_test_user, skip_confirmation=True)
    deleter.delete_label_and_samples("Nice")
    time.sleep(2)

    samples = func.list_samples()
    assert len(samples) == 3
    for sample in samples:
        if sample.annotation:
            assert "Nice" not in sample.annotation.label_name

    labels = func.list_labels()
    assert len(labels) == 1



from typing import Dict, List

from nyckel.auth import OAuth2Renewer
from nyckel.functions.classification import factory
from nyckel.functions.classification.classification import (
    ClassificationAnnotation,
    ClassificationFunction,
    ClassificationFunctionURLHandler,
    ClassificationLabel,
    ClassificationPrediction,
    TextClassificationSample,
)
from nyckel.functions.classification.function_handler import ClassificationFunctionHandler
from nyckel.functions.classification.label_handler import ClassificationLabelHandler
from nyckel.functions.classification.sample_handler import ClassificationSampleHandler
from nyckel.functions.utils import strip_nyckel_prefix
from nyckel.request_utils import get_session_that_retries


class TextClassificationFunction(ClassificationFunction):
    def __init__(self, function_id: str, auth: OAuth2Renewer):
        self._function_id = function_id
        self._auth = auth

        self._function_handler = ClassificationFunctionHandler(function_id, auth)
        self._label_handler = ClassificationLabelHandler(function_id, auth)
        self._url_handler = ClassificationFunctionURLHandler(function_id, auth.server_url)
        self._sample_handler = ClassificationSampleHandler(function_id, auth)
        self._session = get_session_that_retries()
        self._refresh_auth_token()

        assert self._function_handler.get_input_modality() == "Text"

    @classmethod
    def new(cls, name: str, auth: OAuth2Renewer) -> "TextClassificationFunction":
        return factory.ClassificationFunctionFactory.new(name, "Text", auth)  # type:ignore

    @classmethod
    def create_function(cls, name: str, auth: OAuth2Renewer) -> "TextClassificationFunction":
        return cls.new(name, auth)

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        status_string = f"Name: {self.get_name()}, id: {self.function_id}, url: {self.train_page}"
        return status_string

    def __call__(self, sample_data: str) -> ClassificationPrediction:
        return self.invoke([sample_data])[0]

    @property
    def function_id(self) -> str:
        return self._function_id

    @property
    def sample_count(self) -> int:
        return self._function_handler.sample_count

    @property
    def label_count(self) -> int:
        return self._function_handler.label_count

    @property
    def train_page(self) -> str:
        return self._url_handler.train_page

    @property
    def input_modality(self) -> str:
        return "Text"

    def get_name(self) -> str:
        return self._function_handler.get_name()

    @property
    def metrics(self) -> Dict:
        return self._function_handler.get_metrics()

    def invoke(self, sample_data_list: List[str]) -> List[ClassificationPrediction]:
        return self._sample_handler.invoke(sample_data_list, lambda x: x)

    def has_trained_model(self) -> bool:
        return self._function_handler.is_trained

    def create_labels(self, labels: List[ClassificationLabel]) -> List[str]:
        return self._label_handler.create_labels(labels)

    def list_labels(self) -> List[ClassificationLabel]:
        return self._label_handler.list_labels(self.label_count)

    def read_label(self, label_id: str) -> ClassificationLabel:
        return self._label_handler.read_label(label_id)

    def update_label(self, label: ClassificationLabel) -> ClassificationLabel:
        return self._label_handler.update_label(label)

    def delete_label(self, label_id: str) -> None:
        return self._label_handler.delete_label(label_id)

    def delete_labels(self, label_ids: List[str]) -> None:
        return self._label_handler.delete_labels(label_ids)

    def create_samples(self, samples: List[TextClassificationSample]) -> List[str]:  # type: ignore
        return self._sample_handler.create_samples(samples, lambda x: x)

    def list_samples(self) -> List[TextClassificationSample]:  # type: ignore
        samples_dict_list = self._sample_handler.list_samples(self.sample_count)
        labels = self._label_handler.list_labels(None)

        label_name_by_id = {label.id: label.name for label in labels}

        return [self._sample_from_dict(entry, label_name_by_id) for entry in samples_dict_list]  # type: ignore

    def read_sample(self, sample_id: str) -> TextClassificationSample:
        sample_dict = self._sample_handler.read_sample(sample_id)

        labels = self._label_handler.list_labels(None)
        label_name_by_id = {strip_nyckel_prefix(label.id): label.name for label in labels}  # type: ignore

        return self._sample_from_dict(sample_dict, label_name_by_id)  # type: ignore

    def update_annotation(self, sample: TextClassificationSample) -> None:  # type: ignore
        self._sample_handler.update_annotation(sample)

    def delete_sample(self, sample_id: str) -> None:
        self._sample_handler.delete_sample(sample_id)

    def delete_samples(self, sample_ids: List[str]) -> None:
        self._sample_handler.delete_samples(sample_ids)

    def delete(self) -> None:
        self._function_handler.delete()

    def _refresh_auth_token(self) -> None:
        self._session.headers.update({"authorization": "Bearer " + self._auth.token})

    def _sample_from_dict(self, sample_dict: Dict, label_name_by_id: Dict[str, str]) -> TextClassificationSample:
        if "annotation" in sample_dict:
            annotation = ClassificationAnnotation(
                label_name=label_name_by_id[strip_nyckel_prefix(sample_dict["annotation"]["labelId"])],
            )
        else:
            annotation = None
        if "prediction" in sample_dict:
            prediction = ClassificationPrediction(
                confidence=sample_dict["prediction"]["confidence"],
                label_name=label_name_by_id[strip_nyckel_prefix(sample_dict["prediction"]["labelId"])],
            )
        else:
            prediction = None
        return TextClassificationSample(
            id=strip_nyckel_prefix(sample_dict["id"]),
            data=sample_dict["data"],
            external_id=sample_dict["externalId"] if "externalId" in sample_dict else None,
            annotation=annotation,
            prediction=prediction,
        )



import sys
import time
from datetime import datetime

import fire

from nyckel import (
    ClassificationFunctionFactory,
    OAuth2Renewer,
)
from nyckel.functions.classification.classification import ClassificationFunction
from nyckel.functions.classification.function_handler import ClassificationFunctionHandler
from nyckel.request_utils import get_session_that_retries


class NyckelFunctionDuplicator:
    def __init__(self, from_function_id: str, auth: OAuth2Renewer, skip_confirmation: bool = False):
        self._from_function_id = from_function_id
        self._auth = auth
        self._skip_confirmation = skip_confirmation
        self._session = get_session_that_retries()
        self._function_handler: ClassificationFunctionHandler

    def __call__(self) -> ClassificationFunction:
        print(f"-> Starting copy process for {self._from_function_id} ...")
        self._refresh_auth_token()
        self._function_handler = ClassificationFunctionHandler(self._from_function_id, self._auth)
        self._validate_function_type()
        self._request_user_confirmation()
        return self._make_copy()

    def _validate_function_type(self) -> None:
        if not self._function_handler.get_output_modality() == "Classification":
            raise ValueError("Can only copy Classification functions.")
        if self._function_handler.get_input_modality() not in ["Text", "Image", "Tabular"]:
            raise ValueError(f"Unknown output type {self._function_handler.get_input_modality()}")

    def _request_user_confirmation(self) -> None:
        if self._skip_confirmation:
            return None
        reply = input(
            f"-> This will copy function: [name: {self._function_handler.get_name()}, id: {self._from_function_id}, "
            f"label-count: {self._function_handler.label_count}, sample-count: {self._function_handler.sample_count}]. "
            f"Ok (y/n)? "
        )
        if not reply == "y":
            print("-> Ok. Aborting...")
            sys.exit(0)

    def _make_copy(self) -> ClassificationFunction:
        print(f"-> Reading labels and samples from {self._from_function_id}")
        from_function = ClassificationFunctionFactory.load(self._from_function_id, self._auth)
        labels = from_function.list_labels()
        samples = from_function.list_samples()

        to_function = ClassificationFunctionFactory.new(
            self.new_function_name, from_function.input_modality, self._auth
        )

        print(f"-> Adding labels and samples to {to_function.function_id}")
        to_function.create_labels(labels)
        to_function.create_samples(samples)
        time.sleep(2)  # Pause to allow assets to be available via the API.
        print(f"-> Done! New function available at: {to_function.train_page}")
        return to_function

    @property
    def new_function_name(self) -> str:
        return f"COPY OF {self._function_handler.get_name()} -- {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"

    def _refresh_auth_token(self) -> None:
        self._session.headers.update({"authorization": "Bearer " + self._auth.token})


def main(client_id: str, client_secret: str, from_function_id: str, server_url: str = "https://www.nyckel.com") -> None:
    auth = OAuth2Renewer(client_id, client_secret, server_url)
    duplicator = NyckelFunctionDuplicator(from_function_id, auth)
    duplicator()


if __name__ == "__main__":
    fire.Fire(main)



import time

from nyckel import OAuth2Renewer, TabularClassificationFunction
from nyckel.functions.classification.tabular_classification import TabularFieldHandler, TabularFunctionField


def test_fields(tabular_classification_function: TabularClassificationFunction, auth_test_user: OAuth2Renewer) -> None:
    func = tabular_classification_function
    # Just borrow the function_id and create a TabularFieldHandler directly to test it.
    fields_handler = TabularFieldHandler(func.function_id, auth_test_user)

    field = TabularFunctionField(name="Firstname", type="Text")
    field_id = fields_handler.create_fields([field])[0]
    fields_back = fields_handler.read_field(field_id)
    assert fields_back.name == "Firstname"
    assert fields_back.type == "Text"

    fields = fields_handler.list_fields()
    if len(fields) < 1:
        # Give the API a chance to catch up.
        time.sleep(1)
        fields = fields_handler.list_fields()
    assert len(fields) == 1

    # And delete label
    fields_handler.delete_field(field_id)
    fields = fields_handler.list_fields()
    if not len(fields) == 0:
        # Give the API a chance to catch up.
        time.sleep(1)
        fields = fields_handler.list_fields()
    assert len(fields) == 0



import numbers
import time
from typing import Dict, List

from tqdm import tqdm

from nyckel import OAuth2Renewer
from nyckel.functions.classification import factory
from nyckel.functions.classification.classification import (
    ClassificationAnnotation,
    ClassificationFunction,
    ClassificationFunctionURLHandler,
    ClassificationLabel,
    ClassificationPrediction,
    TabularClassificationSample,
    TabularFunctionField,
)
from nyckel.functions.classification.function_handler import ClassificationFunctionHandler
from nyckel.functions.classification.label_handler import ClassificationLabelHandler
from nyckel.functions.classification.sample_handler import ClassificationSampleHandler
from nyckel.functions.utils import strip_nyckel_prefix
from nyckel.request_utils import ParallelPoster, SequentialGetter, get_session_that_retries


class TabularClassificationFunction(ClassificationFunction):
    def __init__(self, function_id: str, auth: OAuth2Renewer) -> None:
        self._function_id = function_id
        self._auth = auth

        self._function_handler = ClassificationFunctionHandler(function_id, auth)
        self._label_handler = ClassificationLabelHandler(function_id, auth)
        self._field_handler = TabularFieldHandler(function_id, auth)
        self._sample_handler = ClassificationSampleHandler(function_id, auth)
        self._url_handler = ClassificationFunctionURLHandler(function_id, auth.server_url)
        self._session = get_session_that_retries()
        self._refresh_auth_token()
        assert self._function_handler.get_input_modality() == "Tabular"

    @classmethod
    def new(cls, name: str, auth: OAuth2Renewer) -> "TabularClassificationFunction":
        return factory.ClassificationFunctionFactory.new(name, "Tabular", auth)  # type: ignore

    @classmethod
    def create_function(cls, name: str, auth: OAuth2Renewer) -> "TabularClassificationFunction":
        return cls.new(name, auth)

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        status_string = f"Name: {self.get_name()}, id: {self.function_id}, url: {self.train_page}"
        return status_string

    def __call__(self, sample_data: Dict) -> ClassificationPrediction:  # type: ignore
        return self.invoke([sample_data])[0]

    @property
    def function_id(self) -> str:
        return self._function_id

    @property
    def sample_count(self) -> int:
        return self._function_handler.sample_count

    @property
    def label_count(self) -> int:
        return self._function_handler.label_count

    @property
    def train_page(self) -> str:
        return self._url_handler.train_page

    @property
    def input_modality(self) -> str:
        return "Tabular"

    def get_name(self) -> str:
        return self._function_handler.get_name()

    @property
    def metrics(self) -> Dict:
        return self._function_handler.get_metrics()

    def invoke(self, sample_data_list: List[Dict]) -> List[ClassificationPrediction]:  # type: ignore
        return self._sample_handler.invoke(sample_data_list, lambda x: x)

    def has_trained_model(self) -> bool:
        return self._function_handler.is_trained

    def create_labels(self, labels: List[ClassificationLabel]) -> List[str]:
        return self._label_handler.create_labels(labels)

    def list_labels(self) -> List[ClassificationLabel]:
        return self._label_handler.list_labels(self.label_count)

    def read_label(self, label_id: str) -> ClassificationLabel:
        return self._label_handler.read_label(label_id)

    def update_label(self, label: ClassificationLabel) -> ClassificationLabel:
        return self._label_handler.update_label(label)

    def delete_label(self, label_id: str) -> None:
        return self._label_handler.delete_label(label_id)

    def delete_labels(self, label_ids: List[str]) -> None:
        return self._label_handler.delete_labels(label_ids)

    def create_samples(self, samples: List[TabularClassificationSample]) -> List[str]:  # type: ignore # noqa: E501
        self._create_fields_as_needed(samples)
        existing_fields = self._field_handler.list_fields()
        field_id_by_name = {field.name: field.id for field in existing_fields}
        samples = [self._switch_field_names_to_field_ids(sample, field_id_by_name) for sample in samples]  # type: ignore # noqa: E501
        return self._sample_handler.create_samples(samples, lambda x: x)

    def _switch_field_names_to_field_ids(
        self, sample: TabularClassificationSample, field_id_by_name: Dict[str, str]
    ) -> TabularClassificationSample:
        field_names = list(sample.data.keys())
        for field_name in field_names:
            field_value = sample.data.pop(field_name)
            sample.data[field_id_by_name[field_name]] = field_value
        return sample

    def list_samples(self) -> List[TabularClassificationSample]:  # type: ignore
        samples_dict_list = self._sample_handler.list_samples(self.sample_count)
        labels = self._label_handler.list_labels(None)
        fields = self._field_handler.list_fields()

        label_name_by_id = {label.id: label.name for label in labels}
        field_name_by_id = {field.id: field.name for field in fields}  # type: ignore

        return [self._sample_from_dict(entry, label_name_by_id, field_name_by_id) for entry in samples_dict_list]  # type: ignore # noqa: E501

    def read_sample(self, sample_id: str) -> TabularClassificationSample:
        sample_as_dict = self._sample_handler.read_sample(sample_id)

        labels = self._label_handler.list_labels(None)
        fields = self._field_handler.list_fields()

        label_name_by_id = {label.id: label.name for label in labels}
        field_name_by_id = {field.id: field.name for field in fields}  # type: ignore

        return self._sample_from_dict(sample_as_dict, label_name_by_id, field_name_by_id)  # type: ignore

    def update_annotation(self, sample: TabularClassificationSample) -> None:  # type: ignore
        self._sample_handler.update_annotation(sample)

    def delete_sample(self, sample_id: str) -> None:
        self._sample_handler.delete_sample(sample_id)

    def delete_samples(self, sample_ids: List[str]) -> None:
        self._sample_handler.delete_samples(sample_ids)

    def delete(self) -> None:
        self._function_handler.delete()

    def _create_fields_as_needed(self, samples: List[TabularClassificationSample]) -> None:
        existing_fields = self._field_handler.list_fields()

        new_fields: List[TabularFunctionField] = []
        for sample in samples:
            for field_name, field_value in sample.data.items():
                if field_name in [field.name for field in existing_fields + new_fields]:
                    continue
                if isinstance(field_value, numbers.Number):
                    new_fields.append(TabularFunctionField(name=field_name, type="Number"))
                else:
                    new_fields.append(TabularFunctionField(name=field_name, type="Text"))

        if len(new_fields) > 0:
            self._field_handler.create_fields(new_fields)

    def _sample_from_dict(
        self, sample_dict: Dict, label_name_by_id: Dict[str, str], field_name_by_id: Dict[str, str]
    ) -> TabularClassificationSample:
        tabular_data_body = {
            field_name_by_id[strip_nyckel_prefix(field_id)]: field_data
            for field_id, field_data in sample_dict["data"].items()
        }

        if "externalId" in sample_dict:
            external_id = sample_dict["externalId"]
        else:
            external_id = None

        if "annotation" in sample_dict:
            annotation = ClassificationAnnotation(
                label_name=label_name_by_id[strip_nyckel_prefix(sample_dict["annotation"]["labelId"])],
            )
        else:
            annotation = None

        if "prediction" in sample_dict:
            prediction = ClassificationPrediction(
                confidence=sample_dict["prediction"]["confidence"],
                label_name=label_name_by_id[strip_nyckel_prefix(sample_dict["prediction"]["labelId"])],
            )
        else:
            prediction = None

        return TabularClassificationSample(
            id=strip_nyckel_prefix(sample_dict["id"]),
            data=tabular_data_body,
            external_id=external_id,
            annotation=annotation,
            prediction=prediction,
        )

    def _refresh_auth_token(self) -> None:
        self._session.headers.update({"authorization": "Bearer " + self._auth.token})


class TabularFieldHandler:
    def __init__(self, function_id: str, auth: OAuth2Renewer):
        self._function_id = function_id
        self._auth = auth
        self._url_handler = ClassificationFunctionURLHandler(function_id, auth.server_url)
        self._session = get_session_that_retries()
        self._field_name_by_id: Dict = {}

    def _refresh_auth_token(self) -> None:
        self._session.headers.update({"authorization": "Bearer " + self._auth.token})

    def create_fields(self, fields: List[TabularFunctionField]) -> List[str]:
        self._refresh_auth_token()
        bodies = [{"name": field.name, "type": field.type} for field in fields]
        url = self._url_handler.api_endpoint(path="fields")
        responses = ParallelPoster(self._session, url)(bodies)
        field_ids = [strip_nyckel_prefix(resp.json()["id"]) for resp in responses]

        self._confirm_new_fields_available(fields)

        return field_ids

    def _confirm_new_fields_available(self, new_fields: List[TabularFunctionField]) -> None:
        # Before returning, make sure the assets are available via the API.
        new_fields_available_via_api = False
        timeout_seconds = 5
        t0 = time.time()
        while not new_fields_available_via_api:
            if time.time() - t0 > timeout_seconds:
                raise ValueError("Something went wrong when posting fields.")
            time.sleep(0.5)
            fields_retrieved = self.list_fields()
            new_fields_available_via_api = set([label.name for label in new_fields]).issubset(
                [label.name for label in fields_retrieved]
            )

    def list_fields(self) -> List[TabularFunctionField]:
        self._refresh_auth_token()
        fields_dict_list = SequentialGetter(self._session, self._url_handler.api_endpoint(path="fields"))(
            tqdm(ncols=80, desc="Listing fields")
        )
        return [self._field_from_dict(entry) for entry in fields_dict_list]

    def read_field(self, field_id: str) -> TabularFunctionField:
        self._refresh_auth_token()
        url = self._url_handler.api_endpoint(path=f"fields/{field_id}")
        response = self._session.get(url)
        if not response.status_code == 200:
            raise RuntimeError(
                f"Unable to fetch field {field_id} from {self._url_handler.train_page} "
                f"{response.text=} {response.status_code=}"
            )
        return self._field_from_dict(response.json())

    def delete_field(self, field_id: str) -> None:
        self._refresh_auth_token()
        response = self._session.delete(self._url_handler.api_endpoint(path=f"fields/{field_id}"))
        assert response.status_code == 200, f"Delete failed with {response.status_code=}, {response.text=}"

    def _field_from_dict(self, field_dict: Dict) -> TabularFunctionField:
        return TabularFunctionField(
            name=field_dict["name"], id=strip_nyckel_prefix(field_dict["id"]), type=field_dict["type"]
        )



from .auth import OAuth2Renewer  # noqa: F401
from .functions.classification.classification import (
    ClassificationAnnotation,  # noqa: F401
    ClassificationFunction,  # noqa: F401
    ClassificationLabel,  # noqa: F401
    ClassificationPrediction,  # noqa: F401
    ClassificationSample,  # noqa: F401
    ImageClassificationSample,  # noqa: F401
    TabularClassificationSample,  # noqa: F401
    TextClassificationSample,  # noqa: F401
)
from .functions.classification.factory import ClassificationFunctionFactory  # noqa: F401
from .functions.classification.image_classification import ImageClassificationFunction  # noqa: F401
from .functions.classification.tabular_classification import TabularClassificationFunction  # noqa: F401
from .functions.classification.text_classification import TextClassificationFunction  # noqa: F401



import sys
from typing import Dict, List

import fire  # type: ignore
from tqdm import tqdm

from nyckel import (
    ClassificationFunctionFactory,
    ClassificationLabel,
    OAuth2Renewer,
)
from nyckel.functions.classification.classification import ClassificationFunctionURLHandler
from nyckel.functions.classification.function_handler import ClassificationFunctionHandler
from nyckel.request_utils import SequentialGetter, get_session_that_retries


class NyckelLabelDeleter:
    def __init__(self, function_id: str, auth: OAuth2Renewer, skip_confirmation: bool = False):
        self._function_id = function_id
        self._auth = auth
        self._skip_confirmation = skip_confirmation
        self._function_handler = ClassificationFunctionHandler(self._function_id, self._auth)
        self._function = ClassificationFunctionFactory.load(function_id, auth)
        self._url_handler = ClassificationFunctionURLHandler(function_id, auth.server_url)
        self._session = get_session_that_retries()

    def delete_label_and_samples(self, label_name: str) -> None:
        self._refresh_auth_token()
        function_labels = self._validate_label_name_exists(label_name)
        self._validate_function_type()
        samples_as_list_of_dicts = self._list_samples_for_label(label_name, function_labels)
        self._request_user_confirmation(len(samples_as_list_of_dicts), label_name)
        self._delete_samples([sample["id"] for sample in samples_as_list_of_dicts])
        self._delete_label(label_name, function_labels)

    def _validate_label_name_exists(self, label_name: str) -> List[ClassificationLabel]:
        labels = self._function.list_labels()
        if label_name not in [label.name for label in labels]:
            raise RuntimeError(f"Label: {label_name} not in function {self._function_id}")
        return labels

    def _validate_function_type(self) -> None:
        if not self._function_handler.get_output_modality() == "Classification":
            raise ValueError("Can only copy Classification functions.")
        if self._function_handler.get_input_modality() not in ["Text", "Image", "Tabular"]:
            raise ValueError(f"Unknown output type {self._function_handler.get_input_modality()}")

    def _request_user_confirmation(self, sample_count: int, label_name: str) -> None:
        if self._skip_confirmation:
            return None
        reply = input(
            f"-> This will delete label: '{label_name}' and the {sample_count} samples annotated as '{label_name}'. "
            f"Proceed (y/n)? "
        )
        if not reply == "y":
            print("-> Ok. Aborting...")
            sys.exit(0)

    def _list_samples_for_label(self, label_name: str, function_labels: List[ClassificationLabel]) -> List[Dict]:
        label_id_by_name = {label.name: label.id for label in function_labels}
        url = self._url_handler.api_endpoint(
            path="samples", query_str=f"annotationLabelId={label_id_by_name[label_name]}"
        )
        samples_list_dict = SequentialGetter(self._session, url)(tqdm(ncols=80, desc="Listing samples"))
        return samples_list_dict

    def _delete_samples(self, sample_ids: List[str]) -> None:
        self._function.delete_samples(sample_ids)
        print(f"-> {len(sample_ids)} samples deleted.")

    def _delete_label(self, label_name: str, function_labels: List[ClassificationLabel]) -> None:
        label_id_by_name = {label.name: label.id for label in function_labels}
        self._function.delete_label(label_id_by_name[label_name])  # type: ignore
        print(f"-> Label '{label_name}' deleted.")

    def _refresh_auth_token(self) -> None:
        self._session.headers.update({"authorization": "Bearer " + self._auth.token})


def main(
    client_id: str, client_secret: str, function_id: str, label_name: str, server_url: str = "https://www.nyckel.com"
) -> None:
    auth = OAuth2Renewer(client_id, client_secret, server_url)
    label_deleter = NyckelLabelDeleter(function_id, auth)
    label_deleter.delete_label_and_samples(label_name)


if __name__ == "__main__":
    fire.Fire(main)



import time

from nyckel import (
    ClassificationLabel,
)
from nyckel.functions.classification.text_classification import TextClassificationFunction


def test_labels(text_classification_function: TextClassificationFunction) -> None:
    # Try creating a simple label
    func = text_classification_function
    label = ClassificationLabel(name="Nice")
    label_ids = func.create_labels([label])
    label_back = func.read_label(label_ids[0])
    assert label_back.name == "Nice"

    # And then one with metadata
    label = ClassificationLabel(name="Nicer", description="Very nice", metadata={"how nice?": "very"})
    label_ids = func.create_labels([label])
    nicer_label_id = label_ids[0]
    label_back = func.read_label(nicer_label_id)
    assert label_back.name == label.name
    assert label_back.description == label.description
    assert label_back.metadata == label.metadata

    # Check that list_labels work
    labels = func.list_labels()
    if len(labels) < 2:
        # Give the API a chance to catch up.
        time.sleep(1)
        labels = func.list_labels()
    assert len(labels) == 2
    assert set([label.name for label in labels]) == set(["Nice", "Nicer"])

    # And delete label
    func.delete_label(nicer_label_id)
    labels = func.list_labels()
    if not len(labels) == 1:
        # Give the API a chance to catch up.
        time.sleep(1)
        labels = func.list_labels()
    assert len(labels) == 1
    assert labels[0].name == "Nice"


def test_parallel_delete(text_classification_function_with_content: TextClassificationFunction) -> None:
    func: TextClassificationFunction = text_classification_function_with_content
    labels = func.list_labels()
    label_ids = [label.id for label in labels if label.id]
    func.delete_labels(label_ids)
    time.sleep(1)
    assert len(func.list_labels()) == 0


def test_labels_optional_params(text_classification_function: TextClassificationFunction) -> None:
    # Try creating a simple label
    func = text_classification_function

    label = ClassificationLabel(name="Nicer", description="Very nice")
    label_id = func.create_labels([label])[0]
    label_back = func.read_label(label_id)
    assert label_back.name == label.name
    assert label_back.description == label.description
    assert label_back.metadata == label.metadata



from typing import Dict

from nyckel.auth import OAuth2Renewer
from nyckel.functions.classification.classification import ClassificationFunctionURLHandler
from nyckel.request_utils import get_session_that_retries


class ClassificationFunctionHandler:
    def __init__(self, function_id: str, auth: OAuth2Renewer):
        self._function_id = function_id
        self._auth = auth
        self._session = get_session_that_retries()
        self._url_handler = ClassificationFunctionURLHandler(function_id, auth.server_url)
        self.validate_access()
        assert self.get_output_modality() == "Classification"

    @property
    def sample_count(self) -> int:
        return self.get_metrics()["sampleCount"]

    @property
    def label_count(self) -> int:
        return len(self.get_metrics()["annotatedLabelCounts"])

    @property
    def is_trained(self) -> bool:
        metrics = self.get_metrics()
        meta = self.get_v09_function_meta()
        return (
            not metrics["isTraining"]
            and metrics["sampleCount"] == metrics["predictionCount"]
            and meta["state"] in ["Browsing", "Tuning"]
        )

    def validate_access(self) -> None:
        self._refresh_auth_token()

        url = f"{self._auth.server_url}/v1/functions/{self._function_id}"
        response = self._session.get(url)

        if response.status_code == 401:
            raise ValueError(f"Invalid access tokens. Can't access {self._function_id}.")
        if response.status_code == 403:
            raise ValueError(
                f"Can't access {self._function_id} using credentials with CLIENT_ID: {self._auth.client_id}."
            )
        elif not response.status_code == 200:
            raise ValueError(
                f"Failed to load function with id = {self._function_id}. Status code: {response.status_code}"
            )

    def get_name(self) -> str:
        self._refresh_auth_token()
        url = f"{self._auth.server_url}/v1/functions/{self._function_id}"
        response = self._session.get(url)
        assert response.status_code == 200
        return response.json()["name"] if "name" in response.json() else "NewFunction"

    def get_metrics(self) -> Dict:
        self._refresh_auth_token()
        url = self._url_handler.api_endpoint(path="metrics", api_version="v0.9")
        resp = self._session.get(url)
        if not resp.status_code == 200:
            raise RuntimeError(f"Can't get {url=}. {resp.status_code=} {resp.text=}")
        return resp.json()

    def get_input_modality(self) -> str:
        self._refresh_auth_token()
        url = self._url_handler.api_endpoint()
        response = self._session.get(url)
        assert response.status_code == 200, f"{response.text=} {response.status_code=}"
        return response.json()["input"]

    def get_output_modality(self) -> str:
        self._refresh_auth_token()
        url = self._url_handler.api_endpoint()
        response = self._session.get(url)
        assert response.status_code == 200, f"{response.text=} {response.status_code=}"
        return response.json()["output"]

    def get_v09_function_meta(self) -> Dict:
        self._refresh_auth_token()
        url = self._url_handler.api_endpoint(api_version="v0.9")
        resp = self._session.get(url)
        if not resp.status_code == 200:
            raise RuntimeError(f"Can't get {url=}. {resp.status_code=} {resp.text=}")
        return resp.json()

    def delete(self) -> None:
        self._refresh_auth_token()
        endpoint = self._url_handler.api_endpoint()
        response = self._session.delete(endpoint)
        assert response.status_code == 200, f"Delete failed with {response.status_code=}, {response.text=}"
        print(f"-> Function {self._url_handler.train_page} deleted.")

    def _refresh_auth_token(self) -> None:
        self._session.headers.update({"authorization": "Bearer " + self._auth.token})
